#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING, Awaitable

from ..Message import Message

if TYPE_CHECKING:
    from ..Serializer import Serializer
    from ..TelepactSchema import TelepactSchema


async def process_bytes(request_message_bytes: bytes, override_headers: dict[str, object],
                        serializer: 'Serializer', telepact_schema: 'TelepactSchema',
                        on_error: Callable[[Exception], None], on_request: Callable[['Message'], None],
                        on_response: Callable[['Message'], None], handler: Callable[['Message'], Awaitable['Message']]) -> bytes:
    from ..internal.HandleMessage import handle_message
    from ..internal.ParseRequestMessage import parse_request_message

    try:
        request_message = parse_request_message(
            request_message_bytes, serializer, telepact_schema, on_error)

        try:
            on_request(request_message)
        except Exception:
            pass

        response_message = await handle_message(
            request_message, override_headers, telepact_schema, handler, on_error)

        try:
            on_response(response_message)
        except Exception:
            pass

        return serializer.serialize(response_message)
    except Exception as e:
        try:
            on_error(e)
        except Exception:
            pass

        return serializer.serialize(Message({}, {"ErrorUnknown_": {}}))



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..types.TUnion import TUnion
    from ..types.TStruct import TStruct
    from ..types.TType import TType
    from ..types.TFieldDeclaration import TFieldDeclaration
    from ..types.TTypeDeclaration import TTypeDeclaration


def derive_possible_select(fn_name: str, result: 'TUnion') -> dict[str, object]:
    from ..types.TUnion import TUnion
    from ..types.TStruct import TStruct

    nested_types: dict[str, TType] = {}
    ok_fields: dict[str, TFieldDeclaration] = result.tags['Ok_'].fields

    ok_field_names = sorted(ok_fields.keys())

    for field_decl in ok_fields.values():
        find_nested_types(field_decl.type_declaration, nested_types)

    possible_select: dict[str, object] = {}

    possible_select['->'] = {
        'Ok_': ok_field_names,
    }

    sorted_type_keys = sorted(nested_types.keys())
    for k in sorted_type_keys:
        if k.startswith('fn.'):
            continue

        v = nested_types[k]
        if isinstance(v, TUnion):
            union_select: dict[str, list[str]] = {}
            sorted_tag_keys = sorted(v.tags.keys())
            for c in sorted_tag_keys:
                typ = v.tags[c]
                selected_field_names: list[str] = sorted(typ.fields.keys())

                if len(selected_field_names) > 0:
                    union_select[c] = selected_field_names

            possible_select[k] = union_select
        elif isinstance(v, TStruct):
            struct_select: list[str] = sorted(v.fields.keys())

            if len(struct_select) > 0:
                possible_select[k] = struct_select

    return possible_select


def find_nested_types(type_declaration: 'TTypeDeclaration', nested_types: dict[str, 'TType']) -> None:
    from ..types.TUnion import TUnion
    from ..types.TStruct import TStruct
    from ..types.TArray import TArray
    from ..types.TObject import TObject

    typ = type_declaration.type

    if isinstance(typ, TUnion):
        nested_types[typ.name] = typ
        for tag in typ.tags.values():
            for field_decl in tag.fields.values():
                find_nested_types(field_decl.type_declaration, nested_types)

    elif isinstance(typ, TStruct):
        nested_types[typ.name] = typ
        for field_decl in typ.fields.values():
            find_nested_types(field_decl.type_declaration, nested_types)
    
    elif isinstance(typ, TArray) or isinstance(typ, TObject):
        find_nested_types(type_declaration.type_parameters[0], nested_types)


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
import base64

from ...internal.types.TBytes import _BYTES_NAME

if TYPE_CHECKING:
    from .ValidationFailure import ValidationFailure
    from .ValidateContext import ValidateContext


def validate_bytes(value: object, ctx: 'ValidateContext') -> list['ValidationFailure']:
    from .GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure

    if isinstance(value, bytes):
        if ctx.coerce_base64:
            set_coerced_path(ctx.path, ctx.base64_coercions)
        return []
    if isinstance(value, str):
        try:
            base64.b64decode(value)
            if not ctx.coerce_base64:
                set_coerced_path(ctx.path, ctx.bytes_coercions)
            return []
        except Exception as e:
            return get_type_unexpected_validation_failure([], value, 'Base64String')
    else:
        return get_type_unexpected_validation_failure([], value, _BYTES_NAME)
        

def set_coerced_path(path: list[str], coerced_path: dict[str, object]):
    part = path[0]

    if len(path) > 1:
        coerced_path[part] = coerced_path.get(part, {})
        set_coerced_path(path[1:], cast(dict[str, object], coerced_path[part]))
    else:
        coerced_path[part] = True



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ..validation.ValidateContext import ValidateContext
    from ...RandomGenerator import RandomGenerator
    from .TTypeDeclaration import TTypeDeclaration
    from ..validation.ValidationFailure import ValidationFailure
    from ..generation.GenerateContext import GenerateContext


from .TType import TType

_NUMBER_NAME: str = "Number"


class TNumber(TType):

    def get_type_parameter_count(self) -> int:
        return 0

    def validate(self, value: object,
                 type_parameters: list['TTypeDeclaration'], ctx: 'ValidateContext') -> list['ValidationFailure']:
        from ..validation.ValidateNumber import validate_number
        return validate_number(value)

    def generate_random_value(self, blueprint_value: object, use_blueprint_value: bool, type_parameters: list['TTypeDeclaration'], ctx: 'GenerateContext') -> object:
        from ..generation.GenerateRandomNumber import generate_random_number
        return generate_random_number(blueprint_value, use_blueprint_value, ctx)

    def get_name(self, ctx: 'ValidateContext') -> str:
        return _NUMBER_NAME



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast
from msgpack import ExtType

from ...internal.binary.PackList import PACKED_BYTE
from ...internal.binary.Unpack import unpack
from ...internal.binary.UnpackMap import unpack_map


def unpack_list(lst: list[object]) -> list[object]:
    if not lst:
        return lst

    if not isinstance(lst[0], ExtType) or lst[0].code != PACKED_BYTE:
        new_lst = []
        for item in lst:
            new_lst.append(unpack(item))
        return new_lst

    unpacked_lst: list[object] = []
    headers = cast(list[object], lst[1])

    for i in range(2, len(lst)):
        row = cast(list[object], lst[i])
        m = unpack_map(row, headers)

        unpacked_lst.append(m)

    return unpacked_lst



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...RandomGenerator import RandomGenerator


class GenerateContext:
    include_optional_fields: bool
    randomize_optional_fields: bool
    always_include_required_fields: bool
    fn_scope: str
    random_generator: 'RandomGenerator'

    def __init__(self, include_optional_fields: bool,
                 randomize_optional_fields: bool, always_include_required_fields: bool,
                 fn_scope: str, random_generator: 'RandomGenerator') -> None:
        self.include_optional_fields = include_optional_fields
        self.randomize_optional_fields = randomize_optional_fields
        self.always_include_required_fields = always_include_required_fields
        self.fn_scope = fn_scope
        self.random_generator = random_generator

    def copy(self, include_optional_fields: bool | None = None,
             randomize_optional_fields: bool | None = None, always_include_required_fields: bool | None = None,
             fn_scope: str | None = None, random_generator: 'RandomGenerator | None' = None) -> 'GenerateContext':
        return GenerateContext(include_optional_fields if include_optional_fields is not None else self.include_optional_fields,
                               randomize_optional_fields if randomize_optional_fields is not None else self.randomize_optional_fields,
                               always_include_required_fields if always_include_required_fields is not None else self.always_include_required_fields,
                               fn_scope if fn_scope is not None else self.fn_scope,
                               random_generator if random_generator is not None else self.random_generator)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast


if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ..types.TStruct import TStruct
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.generation.GenerateContext import GenerateContext


def generate_random_fn(blueprint_value: object, use_blueprint_value: bool, call_tags: dict[str, 'TStruct'], ctx: 'GenerateContext') -> object:
    from ...internal.generation.GenerateRandomUnion import generate_random_union

    return generate_random_union(blueprint_value, use_blueprint_value, call_tags, ctx)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .TTypeDeclaration import TTypeDeclaration


class TFieldDeclaration:
    def __init__(
            self,
            field_name: str,
            type_declaration: 'TTypeDeclaration',
            optional: bool
    ) -> None:
        self.field_name = field_name
        self.type_declaration = type_declaration
        self.optional = optional



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast

from .GenerateContext import GenerateContext
from .GenerateRandomUnion import generate_random_union
from ..types.TType import TType


def generate_random_mock_call(types: dict[str, TType], ctx: GenerateContext) -> object:
    from ..types.TUnion import TUnion

    function_names = [
        key for key in types.keys()
        if key.startswith('fn.') and not key.endswith('.->') and not key.endswith('_')
    ]

    function_names.sort()

    selected_fn_name = function_names[ctx.random_generator.next_int_with_ceiling(
        len(function_names))]
    
    selected_fn = cast(TUnion, types[selected_fn_name])

    return generate_random_union(None, False, selected_fn.tags, ctx.copy(always_include_required_fields=False))



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from abc import ABCMeta, abstractmethod
from typing import Any

from .BinaryEncoding import BinaryEncoding

class BinaryEncodingCache(metaclass=ABCMeta):

    @abstractmethod
    def add(self, checksum: int, binary_encoding_map: dict[str, int]) -> None:
        """
        Set a binary encoding in the cache.

        Args:
            binary_encoding: The binary encoding.
            checksum: The checksum of the binary encoding.
        """
        pass

    @abstractmethod
    def get(self, checksum: int) -> 'BinaryEncoding':
        """
        Get a binary encoding from the cache.

        Args:
            checksum: The checksum of the binary encoding.

        Returns:
            The binary encoding.
        """
        pass

    @abstractmethod
    def remove(self, checksum: int) -> None:
        """
        Delete a binary encoding from the cache.

        Args:
            checksum: The checksum of the binary encoding.
        """
        pass


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

from ..internal.binary.BinaryEncoderUnavailableError import BinaryEncoderUnavailableError
from ..SerializationError import SerializationError

if TYPE_CHECKING:
    from ..Serialization import Serialization
    from ..Message import Message
    from ..internal.binary.BinaryEncoder import BinaryEncoder
    from ..internal.binary.Base64Encoder import Base64Encoder

def serialize_internal(message: 'Message', binary_encoder: 'BinaryEncoder',
                       base64_encoder: 'Base64Encoder',
                       serializer: 'Serialization') -> bytes:
    headers: dict[str, object] = message.headers

    serialize_as_binary: bool
    if "@binary_" in headers:
        serialize_as_binary = headers.pop("@binary_") is True
    else:
        serialize_as_binary = False

    message_as_pseudo_json: list[object] = [
        message.headers, message.body]

    try:
        if serialize_as_binary:
            try:
                encoded_message = binary_encoder.encode(message_as_pseudo_json)
                return serializer.to_msgpack(encoded_message)
            except BinaryEncoderUnavailableError:
                # We can still submit as json
                base_64_encoded_message = base64_encoder.encode(message_as_pseudo_json)
                return serializer.to_json(base_64_encoded_message)
        else:
            base_64_encoded_message = base64_encoder.encode(message_as_pseudo_json)
            return serializer.to_json(base_64_encoded_message)
    except Exception as e:
        raise SerializationError() from e



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
import base64

def server_base64_encode(message: list[object]) -> None:
    headers = cast(dict[str, object], message[0])
    body = cast(dict[str, object], message[1])

    base64_paths = cast(dict[str, object], headers.get("@base64_", {}))

    travel_base64_encode(body, base64_paths)

def travel_base64_encode(value: object, base64_paths: object) -> object:
    if isinstance(base64_paths, dict):
        for key, val in base64_paths.items():
            if val is True:
                if key == "*" and isinstance(value, list):
                    for i, v in enumerate(value):
                        nv = travel_base64_encode(v, val)
                        value[i] = nv
                elif key == "*" and isinstance(value, dict):
                    for k, v in value.items():
                        nv = travel_base64_encode(v, val)
                        value[k] = nv
                elif isinstance(value, dict):
                    nv = travel_base64_encode(value[key], val)
                    value[key] = nv
                else:
                    raise ValueError(f"Invalid base64 path: {key} for value: {value}")
            else:
                if key == "*" and isinstance(value, list):
                    for i, v in enumerate(value):
                        travel_base64_encode(v, val)
                elif key == "*" and isinstance(value, dict):
                    for k, v in value.items():
                        travel_base64_encode(v, val)
                elif isinstance(value, dict):
                    travel_base64_encode(value[key], val)
                else:
                    raise ValueError(f"Invalid base64 path: {key} for value: {value}")
        return None
    elif base64_paths is True and (isinstance(value, bytes) or value is None):
        if value is None:
            return None
        return base64.b64encode(value).decode("utf-8")
    else:
        raise ValueError(f"Invalid base64 path: {base64_paths} for value: {value}")



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ..types.TStruct import _STRUCT_NAME

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TFieldDeclaration import TFieldDeclaration
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.validation.ValidationFailure import ValidationFailure


def validate_struct(value: object,
                    name: str, fields: dict[str, 'TFieldDeclaration'], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure
    from ...internal.validation.ValidateStructFields import validate_struct_fields

    if isinstance(value, dict):
        selected_fields = cast(
            list[str], ctx.select.get(name) if ctx.select else None)
        return validate_struct_fields(fields, selected_fields, value, ctx)
    else:
        return get_type_unexpected_validation_failure([], value, _STRUCT_NAME)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast

from ...TelepactSchema import TelepactSchema
from .SchemaParseFailure import SchemaParseFailure

import json


if TYPE_CHECKING:
    from ..types.TType import TType
    from ..types.TFieldDeclaration import TFieldDeclaration


def parse_telepact_schema(
    telepact_schema_document_names_to_json: dict[str, str],
) -> 'TelepactSchema':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from .ApplyErrorToParsedTypes import apply_error_to_parsed_types
    from .CatchErrorCollisions import catch_error_collisions
    from .FindMatchingSchemaKey import find_matching_schema_key
    from .FindSchemaKey import find_schema_key
    from .GetOrParseType import get_or_parse_type
    from .GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from .ParseErrorType import parse_error_type
    from .ParseHeadersType import parse_headers_type
    from .ParseContext import ParseContext
    from .GetPathDocumentCoordinatesPseudoJson import get_path_document_coordinates_pseudo_json
    from ..types.TError import TError
    from collections import OrderedDict
    from ..types.THeaders import THeaders
    from .CatchHeaderCollisions import catch_header_collisions

    original_schema: dict[str, object] = {}
    parsed_types: dict[str, TType] = {}
    fn_error_regexes: dict[str, str] = {}
    parse_failures: list[SchemaParseFailure] = []
    failed_types: set[str] = set()
    schema_keys_to_document_names: dict[str, str] = {}
    schema_keys_to_index: dict[str, int] = {}
    schema_keys: set[str] = set()

    ordered_document_names = sorted(
        list(telepact_schema_document_names_to_json.keys()))

    telepact_schema_document_name_to_pseudo_json: dict[str, list[object]] = {}

    for document_name, telepact_schema_json in telepact_schema_document_names_to_json.items():
        try:
            telepact_schema_pseudo_json_init = json.loads(telepact_schema_json)
        except json.JSONDecodeError as e:
            raise TelepactSchemaParseError(
                [SchemaParseFailure(document_name, [], "JsonInvalid", {})], telepact_schema_document_names_to_json) from e

        if not isinstance(telepact_schema_pseudo_json_init, list):
            this_parse_failure = get_type_unexpected_parse_failure(
                document_name, [], telepact_schema_pseudo_json_init, "Array"
            )
            raise TelepactSchemaParseError(
                this_parse_failure, telepact_schema_document_names_to_json)

        telepact_schema_document_name_to_pseudo_json[document_name] = telepact_schema_pseudo_json_init

    for document_name in ordered_document_names:
        telepact_schema_pseudo_json = telepact_schema_document_name_to_pseudo_json[document_name]

        index = -1
        for definition in telepact_schema_pseudo_json:

            index += 1
            loop_path = [index]

            if not isinstance(definition, dict):
                this_parse_failures = get_type_unexpected_parse_failure(
                    document_name, cast(list[object], loop_path), definition, "Object")
                parse_failures.extend(this_parse_failures)
                continue

            def_ = definition

            try:
                schema_key = find_schema_key(
                    document_name, def_, index, telepact_schema_document_names_to_json)

                matching_schema_key = find_matching_schema_key(
                    schema_keys, schema_key)
                if matching_schema_key is not None:
                    other_path_index = schema_keys_to_index[matching_schema_key]
                    other_document_name = schema_keys_to_document_names[matching_schema_key]
                    final_path = loop_path + [schema_key]
                    final_other_path = [other_path_index, matching_schema_key]
                    document_json = telepact_schema_document_names_to_json[other_document_name]
                    other_location_pseudo_json = get_path_document_coordinates_pseudo_json(
                        final_other_path, document_json)
                    parse_failures.append(
                        SchemaParseFailure(
                            document_name, cast(list[object], final_path),
                            "PathCollision",
                            {
                                "document": other_document_name,
                                "path": final_other_path,
                                "location": other_location_pseudo_json})
                    )
                    continue

                schema_keys.add(schema_key)
                schema_keys_to_index[schema_key] = index
                schema_keys_to_document_names[schema_key] = document_name
                if document_name == 'auto_' or document_name == 'auth_' or not document_name.endswith('_'):
                    original_schema[schema_key] = def_

            except TelepactSchemaParseError as e:
                parse_failures.extend(e.schema_parse_failures)
                continue

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    header_keys: set[str] = set()
    error_keys: set[str] = set()

    for schema_key in schema_keys:
        if schema_key.startswith("info."):
            continue
        elif schema_key.startswith("headers."):
            header_keys.add(schema_key)
            continue
        elif schema_key.startswith("errors."):
            error_keys.add(schema_key)
            continue

        this_index = schema_keys_to_index[schema_key]
        document_name = schema_keys_to_document_names[schema_key]

        try:
            get_or_parse_type(
                [this_index],
                schema_key,
                ParseContext(
                    document_name,
                    telepact_schema_document_name_to_pseudo_json,
                    telepact_schema_document_names_to_json,
                    schema_keys_to_document_names,
                    schema_keys_to_index,
                    parsed_types,
                    fn_error_regexes,
                    parse_failures,
                    failed_types
                )
            )
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    errors: list[TError] = []

    for this_key in error_keys:
        this_index = schema_keys_to_index[this_key]
        this_document_name = schema_keys_to_document_names[this_key]
        telepact_schema_pseudo_json = telepact_schema_document_name_to_pseudo_json[
            this_document_name]
        def_ = cast(dict[str, object],
                    telepact_schema_pseudo_json[this_index])

        try:
            error = parse_error_type(
                [this_index],
                def_,
                this_key,
                ParseContext(
                    this_document_name,
                    telepact_schema_document_name_to_pseudo_json,
                    telepact_schema_document_names_to_json,
                    schema_keys_to_document_names,
                    schema_keys_to_index,
                    parsed_types,
                    fn_error_regexes,
                    parse_failures,
                    failed_types
                )
            )
            errors.append(error)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    try:
        catch_error_collisions(telepact_schema_document_name_to_pseudo_json,
                               error_keys, schema_keys_to_index, schema_keys_to_document_names, telepact_schema_document_names_to_json)
    except TelepactSchemaParseError as e:
        parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    for error in errors:
        try:
            apply_error_to_parsed_types(
                error, parsed_types, schema_keys_to_document_names, schema_keys_to_index, telepact_schema_document_names_to_json, fn_error_regexes)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)

    headers: list[THeaders] = []

    for header_key in header_keys:
        this_index = schema_keys_to_index[header_key]
        this_document_name = schema_keys_to_document_names[header_key]
        telepact_schema_pseudo_json = telepact_schema_document_name_to_pseudo_json[
            this_document_name]
        def_ = cast(dict[str, object],
                    telepact_schema_pseudo_json[this_index])
        try:
            header_type = parse_headers_type(
                [this_index],
                def_,
                header_key,
                ParseContext(
                    this_document_name,
                    telepact_schema_document_name_to_pseudo_json,
                    telepact_schema_document_names_to_json,
                    schema_keys_to_document_names,
                    schema_keys_to_index,
                    parsed_types,
                    fn_error_regexes,
                    parse_failures,
                    failed_types
                )
            )
            headers.append(header_type)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    try:
        catch_header_collisions(telepact_schema_document_name_to_pseudo_json, header_keys,
                                schema_keys_to_index, schema_keys_to_document_names, telepact_schema_document_names_to_json)
    except TelepactSchemaParseError as e:
        parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, telepact_schema_document_names_to_json)

    request_headers: dict[str, TFieldDeclaration] = {}
    response_headers: dict[str, TFieldDeclaration] = {}

    for header in headers:
        request_headers.update(header.request_headers)
        response_headers.update(header.response_headers)

    final_original_schema = [original_schema[k]
                             for k in sorted(original_schema.keys(), key=lambda k: (not k.startswith("info."), k))]

    return TelepactSchema(
        final_original_schema,
        parsed_types,
        request_headers,
        response_headers
    )



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast


if TYPE_CHECKING:
    from ...RandomGenerator import RandomGenerator
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ...internal.generation.GenerateContext import GenerateContext


def generate_random_array(blueprint_value: object, use_blueprint_value: bool, type_parameters: list['TTypeDeclaration'], ctx: 'GenerateContext') -> list[object]:
    nested_type_declaration = type_parameters[0]

    if use_blueprint_value:
        starting_array = cast(list[object], blueprint_value)

        array = []
        for starting_array_value in starting_array:
            value = nested_type_declaration.generate_random_value(
                starting_array_value, True, ctx)

            array.append(value)

        return array
    else:
        length = ctx.random_generator.next_collection_length()

        array = []
        for i in range(length):
            value = nested_type_declaration.generate_random_value(
                None, False, ctx)

            array.append(value)

        return array



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
from ...MockTelepactSchema import MockTelepactSchema


def create_mock_telepact_schema_from_file_json_map(json_documents: dict[str, str]) -> 'MockTelepactSchema':
    from .GetMockTelepactJson import get_mock_telepact_json
    from .CreateTelepactSchemaFromFileJsonMap import create_telepact_schema_from_file_json_map

    final_json_documents = json_documents.copy()
    final_json_documents["mock_"] = get_mock_telepact_json()

    telepact_schema = create_telepact_schema_from_file_json_map(
        final_json_documents)

    return MockTelepactSchema(telepact_schema.original, telepact_schema.parsed, telepact_schema.parsed_request_headers, telepact_schema.parsed_response_headers)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.THeaders import THeaders


def parse_headers_type(path: list[object], headers_definition_as_parsed_json: dict[str, object], schema_key: str, ctx: 'ParseContext') -> 'THeaders':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from ..types.TFieldDeclaration import TFieldDeclaration
    from ...internal.schema.ParseStructFields import parse_struct_fields
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ..types.THeaders import THeaders

    parse_failures = []
    request_headers = {}
    response_headers = {}

    request_headers_def = headers_definition_as_parsed_json.get(schema_key)

    this_path = path + [schema_key]

    if not isinstance(request_headers_def, dict):
        branch_parse_failures = get_type_unexpected_parse_failure(
            ctx.document_name,
            this_path,
            request_headers_def,
            'Object',
        )
        parse_failures.extend(branch_parse_failures)
    else:
        try:
            request_fields = parse_struct_fields(
                this_path, request_headers_def, True, ctx)

            # All headers are optional
            for field in request_fields:
                request_fields[field].optional = True

            request_headers.update(request_fields)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)
        except Exception as e:
            raise e

    response_key = '->'
    response_path = path + [response_key]

    if response_key not in headers_definition_as_parsed_json:
        parse_failures.append(
            SchemaParseFailure(ctx.document_name, response_path, 'RequiredObjectKeyMissing', {
                'key': response_key,
            })
        )

    response_headers_def = headers_definition_as_parsed_json.get(response_key)

    if not isinstance(response_headers_def, dict):
        branch_parse_failures = get_type_unexpected_parse_failure(
            ctx.document_name,
            this_path,
            response_headers_def,
            'Object',
        )
        parse_failures.extend(branch_parse_failures)
    else:
        try:
            response_fields = parse_struct_fields(
                response_path, response_headers_def, True, ctx)

            # All headers are optional
            for field in response_fields:
                response_fields[field].optional = True

            response_headers.update(response_fields)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)
        except Exception as e:
            raise e

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    return THeaders(schema_key, request_headers, response_headers)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
from ..types.TArray import _ARRAY_NAME
from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_array(value: object,
                   type_parameters: list['TTypeDeclaration'], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure

    if isinstance(value, list):
        nested_type_declaration = type_parameters[0]

        validation_failures = []
        for i, element in enumerate(value):
            ctx.path.append("*")

            nested_validation_failures = nested_type_declaration.validate(
                element, ctx)
            index = i

            ctx.path.pop()

            nested_validation_failures_with_path = []
            for f in nested_validation_failures:
                final_path = [index] + f.path

                nested_validation_failures_with_path.append(
                    ValidationFailure(final_path, f.reason, f.data))

            validation_failures.extend(nested_validation_failures_with_path)

        return validation_failures
    else:
        return get_type_unexpected_validation_failure([], value, _ARRAY_NAME)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

def is_sub_map_entry_equal(part_value: object, whole_value: object) -> bool:
    from ...internal.mock.IsSubMap import is_sub_map
    from ...internal.mock.PartiallyMatches import partially_matches

    if isinstance(part_value, dict) and isinstance(whole_value, dict):
        return is_sub_map(part_value, whole_value)
    elif isinstance(part_value, list) and isinstance(whole_value, list):
        for part_element in part_value:
            part_matches = partially_matches(whole_value, part_element)
            if not part_matches:
                return False
        return True
    else:
        return part_value == whole_value



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

from msgpack import ExtType

if TYPE_CHECKING:
    from ...internal.binary.BinaryPackNode import BinaryPackNode


PACKED_BYTE = 17


def pack_list(lst: list[object]) -> list[object]:
    from ...internal.binary.Pack import pack
    from ...internal.binary.PackMap import pack_map
    from ...internal.binary.CannotPack import CannotPack

    if not lst:
        return lst

    packed_list: list[object] = []
    header: list[object] = []

    packed_list.append(ExtType(PACKED_BYTE, b''))

    header.append(None)

    packed_list.append(header)

    key_index_map: dict[int, BinaryPackNode] = {}
    try:
        for item in lst:
            if isinstance(item, dict):
                row = pack_map(item, header, key_index_map)
                packed_list.append(row)
            else:
                # This list cannot be packed, abort
                raise CannotPack()
        return packed_list
    except CannotPack:
        new_list = [pack(item) for item in lst]
        return new_list



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import os
from pathlib import Path
from typing import Dict

from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ...TelepactSchemaParseError import TelepactSchemaParseError


def get_schema_file_map(directory: str) -> Dict[str, str]:
    final_json_documents: Dict[str, str] = {}

    schema_parse_failures = []

    try:
        paths = [str(p) for p in Path(directory).rglob('*')]

        for path in paths:
            relative_path = os.path.relpath(path, directory)

            # Check if directory
            # If it is, add a SchemaParseFailure for DirectoryDisallowed
            if os.path.isdir(path):
                schema_parse_failures.append(SchemaParseFailure(relative_path, [], "DirectoryDisallowed", {}))
                final_json_documents[relative_path] = "[]"
                continue    

            with open(path, 'r') as file:
                content = file.read()
            if not relative_path.endswith('.telepact.json'):
                schema_parse_failures.append(SchemaParseFailure(relative_path, [], "FileNamePatternInvalid", {"expected": "*.telepact.json"}))
            final_json_documents[relative_path] = content
    except IOError as e:
        raise RuntimeError(e)

    if schema_parse_failures:
        raise TelepactSchemaParseError(schema_parse_failures, final_json_documents)

    return final_json_documents



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Optional


class InvalidMessage(Exception):
    def __init__(self, cause: Optional[Exception] = None) -> None:
        super().__init__(str(cause) if cause else None)
        self.cause = cause



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from ..types.TNumber import _NUMBER_NAME
from ...internal.validation.ValidationFailure import ValidationFailure


def validate_number(value: object) -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure

    if isinstance(value, (int, float)) and not isinstance(value, (bool, str)):
        if isinstance(value, (int)):
            if (value > 2**63-1 or value < -(2**63)):
                return [ValidationFailure([], "NumberOutOfRange", {})]
            else:
                return []
        else:
            return []
    else:
        return get_type_unexpected_validation_failure([], value, _NUMBER_NAME)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ...internal.validation.ValidationFailure import ValidationFailure
    from ..types.TType import TType
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_value_of_type(value: object,
                           this_type: 'TType',
                           nullable: bool, type_parameters: list['TTypeDeclaration'],
                           ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure

    if value is None:
        if not nullable:
            return get_type_unexpected_validation_failure([], value, this_type.get_name(ctx))
        else:
            return []

    return this_type.validate(value, type_parameters, ctx)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from .Message import Message
from .Serialization import Serialization
from .internal.binary.BinaryEncoder import BinaryEncoder
from .internal.binary.Base64Encoder import Base64Encoder


class Serializer:
    """
    A serializer that converts a Message to and from a serialized form.
    """

    def __init__(self, serialization_impl: Serialization, binary_encoder: BinaryEncoder, base64_encoder: 'Base64Encoder'):
        self.serialization_impl = serialization_impl
        self.binary_encoder = binary_encoder
        self.base64_encoder = base64_encoder

    def serialize(self, message: Message) -> bytes:
        """
        Serialize a Message into a byte array.
        """
        from .internal.SerializeInternal import serialize_internal
        return serialize_internal(message, self.binary_encoder, self.base64_encoder, self.serialization_impl)

    def deserialize(self, message_bytes: bytes) -> Message:
        """
        Deserialize a Message from a byte array.
        """
        from .internal.DeserializeInternal import deserialize_internal
        return deserialize_internal(message_bytes, self.serialization_impl, self.binary_encoder, self.base64_encoder)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from ...internal.schema.SchemaParseFailure import SchemaParseFailure


def get_type_unexpected_parse_failure(document_name: str, path: list[object], value: object, expected_type: str) -> list[SchemaParseFailure]:
    from ...internal.types.GetType import get_type

    actual_type = get_type(value)
    data: dict[str, object] = {
        "actual": {actual_type: {}},
        "expected": {expected_type: {}}
    }
    return [
        SchemaParseFailure(document_name, path, "TypeUnexpected", data)
    ]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ...internal.validation.ValidationFailure import ValidationFailure
import re

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TTypeDeclaration import TTypeDeclaration
    from ..types.TType import TType


def validate_mock_call(given_obj: object, types: dict[str, 'TType'], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure
    from ..types.TUnion import TUnion

    if not isinstance(given_obj, dict):
        return get_type_unexpected_validation_failure([], given_obj, "Object")

    given_map = given_obj

    regex_string = "^fn\\..*$"

    keys = sorted(given_map.keys())

    matches = [k for k in keys if re.match(regex_string, k)]
    if len(matches) != 1:
        return [ValidationFailure([], "ObjectKeyRegexMatchCountUnexpected",
                                  {"regex": regex_string, "actual": len(matches), "expected": 1, "keys": keys})]

    function_name = matches[0]
    function_def = cast(TUnion, types[function_name])
    input = given_map[function_name]

    function_def_call = function_def
    function_def_name = function_name
    function_def_call_tags = function_def_call.tags

    input_failures = function_def_call_tags[function_def_name].validate(
        input, [], ctx)

    input_failures_with_path = []
    for f in input_failures:
        new_path = [function_name] + f.path

        input_failures_with_path.append(
            ValidationFailure(new_path, f.reason, f.data))

    return [f for f in input_failures_with_path if f.reason != "RequiredObjectKeyMissing"]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast, TYPE_CHECKING

if TYPE_CHECKING:
    from .ClientBinaryStrategy import ClientBinaryStrategy
    from .BinaryEncodingCache import BinaryEncodingCache


def client_binary_decode(message: list[object], binary_encoding_cache: 'BinaryEncodingCache',
                         binary_checksum_strategy: 'ClientBinaryStrategy',
                         ) -> list[object]:
    from ...internal.binary.DecodeBody import decode_body
    from ...internal.binary.UnpackBody import unpack_body

    headers = cast(dict[str, object], message[0])
    encoded_message_body = cast(dict[object, object], message[1])
    binary_checksums = cast(list[int], headers.get("@bin_", []))
    binary_checksum = binary_checksums[0]

    # If there is a binary encoding included on this message, cache it
    if "@enc_" in headers:
        binary_encoding = cast(dict[str, int], headers["@enc_"])
        binary_encoding_cache.add(binary_checksum, binary_encoding)

    binary_checksum_strategy.update_checksum(binary_checksum)
    new_current_checksum_strategy = binary_checksum_strategy.get_current_checksums()

    binary_encoder = binary_encoding_cache.get(new_current_checksum_strategy[0])

    final_encoded_message_body: dict[object, object]
    if headers.get("@pac_") is True:
        final_encoded_message_body = unpack_body(encoded_message_body)
    else:
        final_encoded_message_body = encoded_message_body

    message_body = decode_body(final_encoded_message_body, binary_encoder)
    return [headers, message_body]



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING
import re

if TYPE_CHECKING:
    from ...TelepactSchema import TelepactSchema


def create_telepact_schema_from_file_json_map(json_documents: dict[str, str]) -> 'TelepactSchema':
    from .ParseTelepactSchema import parse_telepact_schema
    from .GetInternalTelepactJson import get_internal_telepact_json
    from .GetAuthTelepactJson import get_auth_telepact_json

    final_json_documents = json_documents.copy()
    final_json_documents["internal_"] = get_internal_telepact_json()

    # Determine if we need to add the auth schema
    for json in json_documents.values():
        regex = re.compile(r'"struct\.Auth_"\s*:')
        matcher = regex.search(json)
        if matcher:
            final_json_documents["auth_"] = get_auth_telepact_json()
            break

    return parse_telepact_schema(final_json_documents)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...internal.validation.ValidationFailure import ValidationFailure


def map_validation_failures_to_invalid_field_cases(
        argument_validation_failures: list['ValidationFailure']) -> list[dict[str, object]]:
    validation_failure_cases: list[dict[str, object]] = []
    for validation_failure in argument_validation_failures:
        validation_failure_case: dict[str, object] = {
            "path": validation_failure.path,
            "reason": {validation_failure.reason: validation_failure.data}
        }
        validation_failure_cases.append(validation_failure_case)

    return validation_failure_cases



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, Optional, cast
import json

if TYPE_CHECKING:
    from .Client import Client
    from .Message import Message
    from .TelepactSchema import TelepactSchema
    from .internal.types.TUnion import TUnion

class TestClient:
    class Options:
        generated_collection_length_min: int = 0
        generated_collection_length_max: int = 3

    def __init__(self, client: 'Client', options: Options):
        from .RandomGenerator import RandomGenerator

        self.client = client
        self.random = RandomGenerator(options.generated_collection_length_min, options.generated_collection_length_max)
        self.schema: Optional['TelepactSchema'] = None

    async def assert_request(self, request_message: 'Message', expected_pseudo_json_body: dict[str, object], expect_match: bool) -> 'Message':
        from .internal.mock.IsSubMap import is_sub_map
        from .Message import Message
        from .TelepactSchema import TelepactSchema
        from .internal.generation.GenerateContext import GenerateContext
        from .internal.types.TUnion import TUnion

        if self.schema is None:
            response = await self.client.request(Message({}, {"fn.api_": {}}))
            ok = cast(dict, response.body["Ok_"])
            api = ok["api"]
            json_str = json.dumps(api)
            self.schema = TelepactSchema.from_json(json_str)

        response_message = await self.client.request(request_message)

        did_match = is_sub_map(expected_pseudo_json_body, response_message.body)

        if expect_match:
            if not did_match:
                raise AssertionError(f"Expected response body was not a sub map. Expected: {expected_pseudo_json_body} Actual: {response_message.body}")
            else:
                return response_message
        else:
            if did_match:
                raise AssertionError(f"Expected response body was a sub map. Expected: {expected_pseudo_json_body} Actual: {response_message.body}")
            else:
                use_blueprint_value = True
                include_optional_fields = False
                always_include_required_fields = True
                randomize_optional_field_generation = False

                function_name = request_message.get_body_target()
                definition = cast(TUnion, self.schema.parsed[f"{function_name}.->"])

                generated_result = definition.generate_random_value(
                    expected_pseudo_json_body, use_blueprint_value, [],
                    GenerateContext(
                        include_optional_fields, randomize_optional_field_generation,
                        always_include_required_fields, function_name,
                        self.random))

                return Message(response_message.headers, cast(dict, generated_result))

    def set_seed(self, seed: int):
        self.random.set_seed(seed)


#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

class ValidateContext:
    path: list[str]
    select: dict[str, object] | None
    fn: str | None
    coerce_base64: bool
    base64_coercions: dict[str, object]
    bytes_coercions: dict[str, object]

    def __init__(self, 
                 select: dict[str, object] | None = None, 
                 fn: str | None = None, 
                 coerce_base64: bool = False):
        self.path = []
        self.select = select
        self.fn = fn
        self.coerce_base64 = coerce_base64
        self.base64_coercions = {}
        self.bytes_coercions = {}
        



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .internal.types.TFieldDeclaration import TFieldDeclaration
    from .internal.types.TType import TType


class MockTelepactSchema:
    """
    A parsed telepact schema.
    """

    def __init__(self, original: list[object], parsed: dict[str, 'TType'], parsed_request_headers: dict[str, 'TFieldDeclaration'],
                 parsed_response_headers: dict[str, 'TFieldDeclaration']):
        self.original = original
        self.parsed = parsed
        self.parsed_request_headers = parsed_request_headers
        self.parsed_response_headers = parsed_response_headers

    @staticmethod
    def from_json(json: str) -> 'MockTelepactSchema':
        from .internal.schema.CreateMockTelepactSchemaFromFileJsonMap import create_mock_telepact_schema_from_file_json_map
        return create_mock_telepact_schema_from_file_json_map({"auto_": json})

    @staticmethod
    def from_file_json_map(file_json_map: dict[str, str]) -> 'MockTelepactSchema':
        from .internal.schema.CreateMockTelepactSchemaFromFileJsonMap import create_mock_telepact_schema_from_file_json_map
        return create_mock_telepact_schema_from_file_json_map(file_json_map)

    @staticmethod
    def from_directory(directory: str) -> 'MockTelepactSchema':
        from .internal.schema.CreateMockTelepactSchemaFromFileJsonMap import create_mock_telepact_schema_from_file_json_map
        from .internal.schema.GetSchemaFileMap import get_schema_file_map
        schema_file_map = get_schema_file_map(directory)
        return create_mock_telepact_schema_from_file_json_map(schema_file_map)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from .Message import Message
    from .MockTelepactSchema import MockTelepactSchema
    from .internal.mock.MockInvocation import MockInvocation
    from .internal.mock.MockStub import MockStub


class MockServer:
    """
    A Mock instance of a telepact server.
    """

    class Options:
        """
        Options for the MockServer.
        """

        def __init__(self) -> None:
            self.on_error: Callable[[Exception], None] = lambda e: None
            self.enable_message_response_generation: bool = True
            self.enable_optional_field_generation: bool = True
            self.randomize_optional_field_generation: bool = True
            self.generated_collection_length_min: int = 0
            self.generated_collection_length_max: int = 3

    def __init__(self, mock_telepact_schema: 'MockTelepactSchema', options: Options) -> None:
        from .Server import Server
        from .RandomGenerator import RandomGenerator
        from .TelepactSchema import TelepactSchema

        self.random: RandomGenerator = RandomGenerator(
            options.generated_collection_length_min, options.generated_collection_length_max)
        self.enableGeneratedDefaultStub: bool = options.enable_message_response_generation
        self.enable_optional_field_generation: bool = options.enable_optional_field_generation
        self.randomize_optional_field_generation: bool = options.randomize_optional_field_generation

        self.stubs: list[MockStub] = []
        self.invocations: list[MockInvocation] = []

        server_options = Server.Options()
        server_options.on_error = options.on_error
        server_options.auth_required = False

        telepact_schema = TelepactSchema(mock_telepact_schema.original, mock_telepact_schema.parsed,
                                       mock_telepact_schema.parsed_request_headers, mock_telepact_schema.parsed_response_headers)

        self.server = Server(
            telepact_schema, self._handle, server_options)

    async def process(self, message: bytes) -> bytes:
        """
        Process a given telepact Request Message into a telepact Response Message.

        :param message: The telepact request message.
        :return: The telepact response message.
        """
        return await self.server.process(message)

    async def _handle(self, request_message: 'Message') -> 'Message':
        from .internal.mock.MockHandle import mock_handle
        return await mock_handle(request_message, self.stubs, self.invocations, self.random,
                                 self.server.telepact_schema, self.enableGeneratedDefaultStub,
                                 self.enable_optional_field_generation, self.randomize_optional_field_generation)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import Tuple, cast, Generator


def get_path_document_coordinates_pseudo_json(path: list[object], document: str) -> dict[str, object]:
    reader = string_reader(document)
    return find_coordinates(path, reader)


def string_reader(s: str) -> Generator[Tuple[str, int, int], str, str]:
    row = 1
    col = 0
    for c in s:
        if c == '\n':
            row += 1
            col = 0
        else:
            col += 1
        yield c, row, col
    return ""


def find_coordinates(path: list[object], reader: Generator[Tuple[str, int, int], str, str], ov_row: int | None = None, ov_col: int | None = None) -> dict[str, object]:
    #print(f"find_coordinates: path={path}")

    for c, row, col in reader:
        if len(path) == 0:
            return {
                'row': ov_row if ov_row else row,
                'col': ov_col if ov_col else col
            }

        #print(f"find_coordinates: char={c}, row={row}, col={col}")
        if c == '{':
            result = find_coordinates_object(path, reader)
            if result:
                return result
        if c == '[':
            result = find_coordinates_array(path, reader)
            if result:
                return result

    raise ValueError("Path not found in document")


def find_coordinates_object(path: list[object], reader: Generator[Tuple[str, int, int], str, str]) -> dict[str, object] | None:
    #print(f"find_coordinates_object: path={path}")
    working_key_row_start = None
    working_key_col_start = None
    for c, row, col in reader:
        #print(f"find_coordinates_object: char={c}, row={row}, col={col}")
        if c == '}':
            return None
        elif c == '"':
            working_key_row_start = row
            working_key_col_start = col
            working_key = find_string(reader)
        elif c == ':':
            if working_key == path[0]:
                return find_coordinates(path[1:], reader, working_key_row_start, working_key_col_start)
            else:
                find_value(reader)

    raise ValueError("Path not found in document")


def find_coordinates_array(path: list[object], reader: Generator[Tuple[str, int, int], str, str]) -> dict[str, object] | None:
    #print(f"find_coordinates_array: path={path}")
    working_index = 0
    if working_index == path[0]:
        return find_coordinates(path[1:], reader)
    else:
        find_value(reader)

    for c, row, col in reader:
        #print(f"find_coordinates_array: char={c}, row={row}, col={col}")
        working_index += 1
        #print(f"find_coordinates_array: working_index={working_index}")
        if working_index == path[0]:
            return find_coordinates(path[1:], reader)
        else:
            find_value(reader)

    raise ValueError("Path not found in document")


def find_value(reader: Generator[Tuple[str, int, int], str, str]) -> bool:
    for c, row, col in reader:
        #print(f"find_value: char={c}, row={row}, col={col}")
        if c == '{':
            find_object(reader)
            return False
        elif c == '[':
            find_array(reader)
            return False
        elif c == '"':
            find_string(reader)
            return False
        elif c == '}':
            return True
        elif c == ']':
            return True
        elif c == ',':
            return False
    raise ValueError("Value not found in document")


def find_object(reader: Generator[Tuple[str, int, int], str, str]) -> None:
    for c, row, col in reader:
        #print(f"find_object: char={c}, row={row}, col={col}")
        if c == '}':
            return
        elif c == '"':
            find_string(reader)
        elif c == ':':
            if find_value(reader):
                return


def find_array(reader: Generator[Tuple[str, int, int], str, str]) -> None:
    #print('find_array')
    if find_value(reader):
        return

    working_index = 0
    for c, row, col in reader:
        #print(f"find_array: char={c}, row={row}, col={col}")
        if c == ']':
            return
        working_index += 1
        if find_value(reader):
            return


def find_string(reader: Generator[Tuple[str, int, int], str, str]) -> str:
    working_string = ""
    for c, row, col in reader:
        #print(f"find_string: char={c}")
        if c == '"':
            return working_string
        else:
            working_string += c
    raise ValueError("String not closed")



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from msgpack import ExtType
from typing import cast

from ...internal.binary.PackMap import UNDEFINED_BYTE


def unpack_map(row: list[object], header: list[object]) -> dict[int, object]:
    from ...internal.binary.Unpack import unpack

    final_map: dict[int, object] = {}

    for j in range(len(row)):
        key = header[j + 1]
        value = row[j]

        if isinstance(value, ExtType) and value.code == UNDEFINED_BYTE:
            continue

        if isinstance(key, list):
            nested_header = key
            nested_row = cast(list[object], value)
            m = unpack_map(nested_row, nested_header)
            i = nested_header[0]

            final_map[i] = m
        else:
            i = key
            unpacked_value = unpack(value)

            final_map[i] = unpacked_value

    return final_map



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING

from ...internal.schema.SchemaParseFailure import SchemaParseFailure

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.TFieldDeclaration import TFieldDeclaration


def parse_struct_fields(path: list[object], reference_struct: dict[str, object], is_header: bool, ctx: 'ParseContext') -> dict[str, 'TFieldDeclaration']:
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.ParseField import parse_field
    from ...internal.schema.GetPathDocumentCoordinatesPseudoJson import get_path_document_coordinates_pseudo_json

    parse_failures = []
    fields: dict[str, 'TFieldDeclaration'] = {}

    for field_declaration, type_declaration_value in reference_struct.items():
        for existing_field in fields.keys():
            existing_field_no_opt = existing_field.split("!")[0]
            field_no_opt = field_declaration.split("!")[0]
            if field_no_opt == existing_field_no_opt:
                final_path = path + [field_declaration]
                final_other_path = path + [existing_field]
                final_other_document_json = ctx.telepact_schema_document_names_to_json[
                    ctx.document_name]
                final_other_location_pseudo_json = get_path_document_coordinates_pseudo_json(
                    final_other_path, final_other_document_json)
                parse_failures.append(SchemaParseFailure(
                    ctx.document_name, final_path, "PathCollision", {
                        "document": ctx.document_name,
                        "path": final_other_path,
                        "location": final_other_location_pseudo_json}))

        try:
            parsed_field = parse_field(path,
                                       field_declaration, type_declaration_value,
                                       is_header, ctx)
            field_name = parsed_field.field_name
            fields[field_name] = parsed_field
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    return fields



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast
import base64


def client_base64_encode(message: list[object]) -> None:
    body = cast(dict[str, object], message[1])

    travel_base64_encode(body)


def travel_base64_encode(value: object) -> None:
    if isinstance(value, dict):
        for key, val in value.items():
            if isinstance(val, bytes):
                value[key] = base64.b64encode(val).decode("utf-8")
            else:
                travel_base64_encode(val)
    elif isinstance(value, list):
        for i, v in enumerate(value):
            if isinstance(v, bytes):
                value[i] = base64.b64encode(v).decode("utf-8")
            else:
                travel_base64_encode(v)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ...internal.schema.SchemaParseFailure import SchemaParseFailure
    from ..types.TType import TType


class ParseContext:
    document_name: str
    telepact_schema_document_names_to_pseudo_json: dict[str, list[object]]
    telepact_schema_document_names_to_json: dict[str, str]
    schema_keys_to_document_name: dict[str, str]
    schema_keys_to_index: dict[str, int]
    parsed_types: dict[str, 'TType']
    fn_error_regexes: dict[str, str]
    all_parse_failures: list['SchemaParseFailure']
    failed_types: set[str]

    def __init__(self, document_name: str, telepact_schema_document_names_to_pseudo_json: dict[str, list[object]],
                 telepact_schema_document_names_to_json: dict[str, str],
                 schema_keys_to_document_name: dict[str, str], schema_keys_to_index: dict[str, int],
                 parsed_types: dict[str, 'TType'], fn_error_regexes: dict[str, str],
                 all_parse_failures: list['SchemaParseFailure'], failed_types: set[str]) -> None:
        self.document_name = document_name
        self.telepact_schema_document_names_to_pseudo_json = telepact_schema_document_names_to_pseudo_json
        self.telepact_schema_document_names_to_json = telepact_schema_document_names_to_json
        self.schema_keys_to_document_name = schema_keys_to_document_name
        self.schema_keys_to_index = schema_keys_to_index
        self.parsed_types = parsed_types
        self.fn_error_regexes = fn_error_regexes
        self.all_parse_failures = all_parse_failures
        self.failed_types = failed_types

    def copy(self, document_name: Optional[str] = None) -> 'ParseContext':
        return ParseContext(document_name if document_name is not None else self.document_name,
                            self.telepact_schema_document_names_to_pseudo_json, self.telepact_schema_document_names_to_json,
                            self.schema_keys_to_document_name, self.schema_keys_to_index,
                            self.parsed_types, self.fn_error_regexes, self.all_parse_failures, self.failed_types)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import TYPE_CHECKING, cast
from ..types.TUnion import _UNION_NAME
from ...internal.validation.ValidationFailure import ValidationFailure

if TYPE_CHECKING:
    from ...internal.validation.ValidateContext import ValidateContext
    from ..types.TStruct import TStruct
    from ..types.TTypeDeclaration import TTypeDeclaration


def validate_union(value: object,
                   name: str, tags: dict[str, 'TStruct'], ctx: 'ValidateContext') -> list['ValidationFailure']:
    from ...internal.validation.GetTypeUnexpectedValidationFailure import get_type_unexpected_validation_failure
    from ...internal.validation.ValidateUnionTags import validate_union_tags

    if isinstance(value, dict):
        selected_tags: dict[str, object]
        if name.startswith("fn."):
            selected_tags = {name: ctx.select.get(
                name) if ctx.select else None}
        else:
            selected_tags = cast(
                dict[str, object], ctx.select.get(name) if ctx.select else None)
        return validate_union_tags(tags, selected_tags, value, ctx)
    else:
        return get_type_unexpected_validation_failure([], value, _UNION_NAME)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

def unpack(value: object) -> object:
    from ...internal.binary.UnpackList import unpack_list

    if isinstance(value, list):
        return unpack_list(value)
    elif isinstance(value, dict):
        new_dict = {}
        for key, val in value.items():
            new_dict[key] = unpack(val)
        return new_dict
    else:
        return value



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

import re
from typing import TYPE_CHECKING
from ...internal.schema.SchemaParseFailure import SchemaParseFailure
from ..types.TUnion import TUnion

if TYPE_CHECKING:
    from ...internal.schema.ParseContext import ParseContext
    from ..types.TType import TType


def parse_union_type(path: list[object], union_definition_as_pseudo_json: dict[str, object], schema_key: str,
                     ignore_keys: list[str], required_keys: list[str],
                     ctx: 'ParseContext') -> 'TUnion':
    from ...TelepactSchemaParseError import TelepactSchemaParseError
    from ...internal.schema.GetTypeUnexpectedParseFailure import get_type_unexpected_parse_failure
    from ...internal.schema.ParseStructFields import parse_struct_fields
    from ..types.TStruct import TStruct

    parse_failures = []

    other_keys = set(union_definition_as_pseudo_json.keys())
    other_keys.discard(schema_key)
    other_keys.discard("///")
    for ignore_key in ignore_keys:
        other_keys.discard(ignore_key)

    if other_keys:
        for k in other_keys:
            loop_path = path + [k]
            parse_failures.append(SchemaParseFailure(
                ctx.document_name, loop_path, "ObjectKeyDisallowed", {}))

    this_path = path + [schema_key]
    def_init = union_definition_as_pseudo_json[schema_key]

    if not isinstance(def_init, list):
        final_parse_failures = get_type_unexpected_parse_failure(
            ctx.document_name, this_path, def_init, "Array")
        parse_failures.extend(final_parse_failures)
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    definition2 = def_init
    definition = []
    index = -1
    for element in definition2:
        index += 1
        loop_path = this_path + [index]
        if not isinstance(element, dict):
            this_parse_failures = get_type_unexpected_parse_failure(
                ctx.document_name, loop_path, element, "Object")
            parse_failures.extend(this_parse_failures)
            continue
        definition.append(element)

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    if len(definition) == 0:
        parse_failures.append(SchemaParseFailure(
            ctx.document_name, this_path, "EmptyArrayDisallowed", {}))
    else:
        for required_key in required_keys:
            for element in definition:
                tag_keys = set(element.keys())
                tag_keys.discard("///")
                if required_key in tag_keys:
                    break
            else:
                branch_path = this_path + [0]
                parse_failures.append(SchemaParseFailure(
                    ctx.document_name, branch_path, "RequiredObjectKeyMissing", {'key': required_key}))

    tags = {}
    tag_indices = {}

    for i, element in enumerate(definition):
        loop_path = this_path + [i]
        map_init = element
        map = dict(map_init)
        map.pop("///", None)
        keys = list(map.keys())

        regex_string = r"^([A-Z][a-zA-Z0-9_]*)$"

        matches = [k for k in keys if re.match(regex_string, k)]
        if len(matches) != 1:
            parse_failures.append(SchemaParseFailure(ctx.document_name, loop_path, "ObjectKeyRegexMatchCountUnexpected",
                                                     {"regex": regex_string, "actual": len(matches),
                                                      "expected": 1, "keys": keys}))
            continue
        if len(map) != 1:
            parse_failures.append(SchemaParseFailure(ctx.document_name, loop_path, "ObjectSizeUnexpected",
                                                     {"expected": 1, "actual": len(map)}))
            continue

        entry = next(iter(map.items()))
        union_tag = entry[0]
        union_key_path = loop_path + [union_tag]

        if not isinstance(entry[1], dict):
            this_parse_failures = get_type_unexpected_parse_failure(
                ctx.document_name, union_key_path, entry[1], "Object")
            parse_failures.extend(this_parse_failures)
            continue
        union_tag_struct = entry[1]

        try:
            fields = parse_struct_fields(union_key_path,
                                         union_tag_struct, False, ctx)
        except TelepactSchemaParseError as e:
            parse_failures.extend(e.schema_parse_failures)
            continue

        union_struct = TStruct(
            f"{schema_key}.{union_tag}", fields)

        tags[union_tag] = union_struct
        tag_indices[union_tag] = i

    if parse_failures:
        raise TelepactSchemaParseError(
            parse_failures, ctx.telepact_schema_document_names_to_json)

    return TUnion(schema_key, tags, tag_indices)



#|
#|  Copyright The Telepact Authors
#|
#|  Licensed under the Apache License, Version 2.0 (the "License");
#|  you may not use this file except in compliance with the License.
#|  You may obtain a copy of the License at
#|
#|  https://www.apache.org/licenses/LICENSE-2.0
#|
#|  Unless required by applicable law or agreed to in writing, software
#|  distributed under the License is distributed on an "AS IS" BASIS,
#|  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#|  See the License for the specific language governing permissions and
#|  limitations under the License.
#|

from typing import cast
from collections import OrderedDict
from ...internal.mock.IsSubMap import is_sub_map
from ...internal.mock.IsSubMapEntryEqual import is_sub_map_entry_equal
from ...internal.mock.MockInvocation import MockInvocation


def verify(function_name: str, argument: dict[str, object], exact_match: bool,
           verification_times: dict[str, object], invocations: list[MockInvocation]) -> dict[str, object]:
    matches_found = 0
    for invocation in invocations:
        if invocation.function_name == function_name:
            if exact_match:
                if invocation.function_argument == argument:
                    invocation.verified = True
                    matches_found += 1
            else:
                is_sub_map_result = is_sub_map(
                    argument, invocation.function_argument)
                if is_sub_map_result:
                    invocation.verified = True
                    matches_found += 1

    all_calls_pseudo_json = [OrderedDict(
        [(invocation.function_name, invocation.function_argument)]) for invocation in invocations]

    verify_times_entry = next(iter(verification_times.items()))
    verify_key = verify_times_entry[0]
    verify_times_struct = cast(dict[str, object], verify_times_entry[1])

    verification_failure_pseudo_json = None
    if verify_key == "Exact":
        times = cast(int, verify_times_struct["times"])
        if matches_found > times:
            verification_failure_pseudo_json = {"TooManyMatchingCalls": OrderedDict([
                ("wanted", {"Exact": {"times": times}}),
                ("found", matches_found),
                ("allCalls", all_calls_pseudo_json)
            ])}
        elif matches_found < times:
            verification_failure_pseudo_json = {"TooFewMatchingCalls": OrderedDict([
                ("wanted", {"Exact": {"times": times}}),
                ("found", matches_found),
                ("allCalls", all_calls_pseudo_json)
            ])}
    elif verify_key == "AtMost":
        times = cast(int, verify_times_struct["times"])
        if matches_found > times:
            verification_failure_pseudo_json = {"TooManyMatchingCalls": OrderedDict([
                ("wanted", {"AtMost": {"times": times}}),
                ("found", matches_found),
                ("allCalls", all_calls_pseudo_json)
            ])}
    elif verify_key == "AtLeast":
        times = cast(int, verify_times_struct["times"])
        if matches_found < times:
            verification_failure_pseudo_json = {"TooFewMatchingCalls": OrderedDict([
                ("wanted", {"AtLeast": {"times": times}}),
                ("found", matches_found),
                ("allCalls", all_calls_pseudo_json)
            ])}

    if verification_failure_pseudo_json is None:
        return {"Ok_": {}}

    return {"ErrorVerificationFailure": {"reason": verification_failure_pseudo_json}}
